"""This is not module that configure library mypythontools, but module that help create config
for your project.

What
====

1) Simple and short syntax.
2) Ability to have docstrings on variables (not dynamically, so visible in IDE) and good for sphinx docs.
3) Type checking and Literal checking via MyProperty.
4) Also function evaluation from other config values (not only static value stored).
5) Options hierarchy (nested options).

Examples:
=========

    >>> from __future__ import annotations
    >>> from typing_extensions import Literal
    ...
    >>> class SimpleConfig(ConfigBase):
    ...     @MyProperty
    ...     @staticmethod
    ...     def var() -> int:  # Type hints are validated.
    ...         '''
    ...         Type:
    ...             int
    ...
    ...         Default:
    ...             123
    ...
    ...         This is docstrings (also visible in IDE, because not defined dynamically).
    ...         Also visible in Sphinx documentation.'''
    ...
    ...         return 123  # This is initial value that can be edited.
    ...
    ...     @MyProperty
    ...     def var_literal(self) -> Literal[1, 2, 3]:  # Literal options are also validated
    ...         return 2
    ...
    ...     @MyProperty   # If other defined value is changed, computed property is also updated
    ...     def evaluated(self) -> int | float:
    ...         return self.var + 1
    ...
    ...
    ...     @property   # If you need some logic in setter, use normal property
    ...     def evaluated(self) -> int | float:
    ...         return self.var + 1
    ...
    >>> config = SimpleConfig()
    >>> config.var
    123
    >>> config.var = 665
    >>> config.var
    665
    >>> config['var']  # You can also access params as in a dictionary
    665
    >>> config.var = "String is problem"  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    TypeError: ...
    ...
    >>> config.var_literal = 4  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    TypeError: ...
    ...
    >>> config.evaluated
    666

    You can still setup a function (or lambda expression) as a new value
    and returned value still will be validated

    >>> config.var = lambda self: self.var_literal + 1

This is how help looks like in VS Code

.. image:: /_static/intellisense.png
    :width: 620
    :alt: tasks
    :align: center


Hierarchical config
-------------------

If last level, do inherit from ConfigBase, else inherit from ConfigStructured

Note:
    Use unique values for all config variables even if they are in various subconfig.

>>> class Config(ConfigStructured):
...     def __init__(self) -> None:
...         self.subconfig1 = self.SubConfiguration1()
...         self.subconfig2 = self.SubConfiguration2()
...
...     class SubConfiguration1(ConfigStructured):
...         def __init__(self) -> None:
...             self.subsubconfig = self.SubSubConfiguration()
...
...         class SubSubConfiguration(ConfigBase):
...             @MyProperty
...             def value1() -> Literal[0, 1, 2, 3]:
...                 '''Documentation here
...
...                 Options: [0, 1, 2, 3]
...                 '''
...                 return 3
...
...             @MyProperty
...             def value2(self):
...                 return self.value1 + 1
...
...     class SubConfiguration2(ConfigBase):
...         @MyProperty
...         def other_val(self):
...             return self.value2 + 1
...
...     # Also subconfig category can contain values itself
...     @MyProperty  # Even if no defining @staticmethod, it will work, but may be confusing
...     def value3() -> int:
...         return 3
...
>>> config = Config()
...
>>> config.subconfig1.subsubconfig.value2
4

You can access value from config as well as from subcategory

>>> config.value2
4

Copy
----

Sometimes you need more instances of settings and you need copy of existing configuration.
Copy is deepcopy by default.

>>> config2 = config.copy()
>>> config2.value3 = 0
>>> config2.value3
0
>>> config.value3
3

Bulk update
-----------

Sometimes you want to update many values with flat dictionary.

>>> config.update({'value3': 2, 'value1': 0})
>>> config.value3
2
>>> config.update({"not_existing": "Should fail"})
Traceback (most recent call last):
AttributeError: ...

Get flat dictionary
-------------------

There is a function that will export all the values to the flat dictionary (no dynamic anymore, just values).

>>> config.get_dict()
{'value3': 2, 'value1': 0, 'value2': 1, 'other_val': 2}

Reset
-----
You can reset to default values

>>> config.value1 = 1
>>> config.reset()
>>> config.value1
3

CLI
---
CLI is provided by argparse. When using `with_argparse` method, it will

1) Create parser and add all arguments with help
2) Parse users' sys args and update config

::

    config.with_argparse()

Now you can use in terminal like.

::

    python my_file.py --value1 12


Only basic types like int, float, str, list, dict, set are possible as eval for using type like numpy
array or pandas dataframe could be security leak.

Sphinx docs
===========

If you want to have documentation via sphinx, you can add this to conf.py::

    napoleon_custom_sections = [
        ("Types", "returns_style"),
        ("Type", "returns_style"),
        ("Options", "returns_style"),
        ("Default", "returns_style"),
        ("For example", "returns_style"),
    ]

Here is example

.. image:: /_static/config_on_sphinx.png
    :width: 620
    :alt: tasks
    :align: center
"""
from __future__ import annotations

from mypythontools.config.config_internal import ConfigBase, ConfigStructured
from mypythontools.property import MyProperty

__all__ = ["ConfigBase", "ConfigStructured", "MyProperty"]



"""Plot data. There is only one main function plot. Check 'plot' documentation for how to use it."""

from mypythontools.plots.plots_internal import get_plotly_layout, plot

__all__ = ["get_plotly_layout", "plot"]



"""
Module where you can get paths used in your project or configure paths to be able other modules here
in mypythontools with PROJECT_PATHS. You can also use `find_path()` to find some path efficiently in
some folder, excluding some other inner folders (like venv, node_modules etc.).There is also function
to get desktop path in posix way.
"""

from mypythontools.paths.paths_internal import (
    find_path,
    get_desktop_path,
    validate_path,
    PathLike,
    PROJECT_PATHS,
    ProjectPaths,
)

__all__ = ["find_path", "get_desktop_path", "validate_path", "PathLike", "PROJECT_PATHS", "ProjectPaths"]



"""Module with functions for 'config' subpackage."""


# TODO - check if some arg name not more times - would be overwritten
# TODO - add frozen and dict_values in __init__
# TODO - Remove one loop over vars (from propagatebase_config_map and do the logic in meta loop)
# TODO - Redefine base_config_map and properties list on class, not in objects in meta

from __future__ import annotations
from typing import Any, TypeVar
from copy import deepcopy
import argparse
import sys

import mylogging

from ..property import init_my_properties, MyProperty  # pylint: disable=unused-import
from .. import misc
from .. import types

ConfigType = TypeVar("ConfigType", bound="ConfigBase")


class ConfigMeta(type):
    """Config metaclass changing init function.

    Main reason is for being able to define own __init__ but
    still has functionality from parent __init__ that is necessary. With this meta, there is no need
    to use super().__init__ by user.

    As user, you probably will not need it. It's used internally if inheriting from ConfigStructured.
    """

    def __init__(cls, name, bases, dct) -> None:
        """Wrap subclass object __init__ to provide Config functionality."""
        type.__init__(cls, name, bases, dct)

        # Avoid base classes here and wrap only user class init
        if name in [
            "ConfigBase",
            "ConfigStructured",
        ]:
            return

        def add_parent__init__(
            self,
            *a,
            dict_values=None,
            frozen=None,
            **kw,
        ):

            self.base_config_map = {}
            self.myproperties_list = []
            self.properties_list = []

            # Call user defined init
            cls._original__init__(self, *a, **kw)

            init_my_properties(self)

            for i, j in vars(type(self)).items():
                if isinstance(j, property):
                    self.properties_list.append(i)

            # If structured config, add proxies to subconfigs
            if self.is_structured:
                self.propagatebase_config_map()

            # Update values from dict param in init
            if dict_values:
                for i, j in dict_values.items():
                    setattr(self, i, j)

            if frozen is None:
                self.frozen = True
            else:
                self.frozen = frozen

        cls._original__init__ = cls.__init__
        cls.__init__ = add_parent__init__  # type: ignore

    def __getitem__(cls, key):
        """To be able to access attributes also on class for example for documentation."""
        return getattr(cls, key)


class ConfigBase(metaclass=ConfigMeta):  # type: ignore
    """Main config class. If need nested config, use ConfigStructured instead.

    You can find working examples in module docstrings.
    """

    # Usually this config is created from someone else that user using this config. Therefore new attributes
    # should not be created. It is possible to force it (raise error). It is possible to set frozen to False
    # to enable creating new instances.
    frozen = False

    is_structured = False

    # base_config_map is used only if using structured config. You can access attribute from subconfig as
    # well as from main config object, there is proxy mapping config dict (same for base and subconfig) if
    # attribute not found on defined object, it will find if it's in this dict where are all config values as
    # key and config object where it's stored. It's poppulated automatically in metaclass during init
    base_config_map = {}

    def __init__(self) -> None:
        """Can be overwritten by user and don't need to be called with super().

        Mainly for linting and type hints.
        """
        self.myproperties_list = []
        self.properties_list = []

    def __new__(cls, *args, **kwargs):
        """Just controll that class is subclassed and not instantiated."""
        if cls is ConfigBase or cls is ConfigStructured:
            raise TypeError(
                "ConfigBase and ConfigStructured are not supposed to be instantiated only to be subclassed."
            )
        return object.__new__(cls, *args, **kwargs)

    def __deepcopy__(self, memo):
        """Provide copy functionality."""
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for i, j in self.__dict__.items():
            setattr(result, i, deepcopy(j, memo))
        return result

    def __getattr__(self, name: str):
        """Controll logic if attribute from other subcongif is used."""
        try:
            # If structured config. Value may not be in this object, but in another related object.
            return getattr(self.base_config_map[name], name)

        except KeyError:

            if name not in [
                "_pytestfixturefunction",
                "__wrapped__",
                "pytest_mock_example_attribute_that_shouldnt_exist",
                "__bases__",
                "__test__",
            ]:

                raise AttributeError(f"Variable {name} not found in config.") from None

    def __setattr__(self, name: str, value: Any) -> None:
        """Setup new config values. Define logic when setting attributes from other subconfig class."""
        setter_name = name[1:] if name.startswith("_") else name

        if (
            not self.frozen
            or name == "frozen"
            or name
            in [
                *self.myproperties_list,
                *self.properties_list,
                *vars(self),
            ]
        ):
            object.__setattr__(self, name, value)

        elif setter_name in self.base_config_map:
            setattr(
                self.base_config_map[setter_name],
                name,
                value,
            )

        else:
            raise AttributeError(
                f"Object {str(self)} is frozen. New attributes cannot be set and attribute '{name}' "
                "not found. Maybe you misspelled name. If you really need to change the value, set "
                "attribute frozen to false."
            )

    def __getitem__(self, key):
        """To be able to be able to use same syntax as if using dictionary."""
        return getattr(self, key)

    def __setitem__(self, key, value):
        """To be able to be able to use same syntax as if using dictionary."""
        setattr(self, key, value)

    def __call__(self, *args: Any, **kwds) -> None:
        """Just to be sure to not be used in unexpected way."""
        raise TypeError("Class is not supposed to be called. Just inherit it to create custom config.")

    def copy(self: ConfigType) -> ConfigType:
        """Create deep copy of config and all it's attributes.

        Returns:
            ConfigType: Deep copy.
        """
        return deepcopy(self)

    def update(self, content: dict) -> None:
        """Bulk update with dict values.

        Args:
            content (dict): E.g {"arg_1": "value"}

        Raises:
            AttributeError: If some arg not found in config.
        """
        for i, j in content.items():
            if hasattr(self, i):
                setattr(self, i, j)
            else:
                raise AttributeError(f"Config has no attribute {i}")

    def reset(self) -> None:
        """Reset config to it's default values."""
        copy = type(self)()

        for i in vars(copy).keys():
            setattr(self, i, copy[i])

        for i in copy.myproperties_list:
            setattr(self, i, copy[i])

        for i in copy.properties_list:
            setattr(self, i, copy[i])

    def get_dict(self) -> dict:
        """Get flat dictionary with it's values.

        Returns:
            dict: Flat config dict.
        """
        normal_vars = {
            key: value
            for key, value in vars(self).items()
            if not key.startswith("__")
            and not callable(value)
            and not hasattr(value, "myproperties_list")
            and key not in ["myproperties_list", "properties_list", "frozen", "base_config_map"]
        }

        property_vars = {
            # Values from myproperties
            **{key: getattr(self, key) for key in self.myproperties_list},
            # Values from properties
            **{key: getattr(self, key) for key in self.properties_list},
        }

        normal_vars = {
            i: j for i, j in normal_vars.items() if not (i.startswith("_") and i[1:] in property_vars)
        }

        return {**normal_vars, **property_vars}

    def with_argparse(self, about: str | None = None) -> None:
        """Parse sys.argv flags and update the config.

        For using with CLI. When using `with_argparse` method.

        1) Create parser and add all arguments with help
        2) Parse users' sys args and update config ::

            config.with_argparse()

        Now you can use in terminal like. ::

            python my_file.py --config_arg config_value

        Only basic types like int, float, str, list, dict, set are possible as eval for using type like numpy
        array or pandas dataframe could be security leak.

        Args:
            about (str, optional): Description used in --help. Defaults to None.

        Raises:
            SystemExit: If arg that do not exists in config.

        Note:
            If using boolean, you must specify the value. Just occurrence, e.g. --my_arg is not True.
        """
        if len(sys.argv) <= 1 or misc.GLOBAL_VARS.jupyter:
            return

        # Add settings from command line if used
        parser = argparse.ArgumentParser(usage=about)

        config_dict = self.get_dict()

        for i in config_dict.keys():
            try:
                help_str = type(self)[i].__doc__
            except AttributeError:
                help_str = type(self.base_config_map[i])[i].__doc__

            parser.add_argument(f"--{i}", help=help_str)

        try:
            parsed_args = parser.parse_known_args()
        except SystemExit as err:
            if err.code == 0:
                sys.exit(0)

            raise SystemExit(
                f"Config args parsing failed. Used args {sys.argv}. Check if args and values are correct "
                "format. Each argument must have just one value. Use double quotes if spaces in string. "
                "For dict e.g. \"{'key': 666}\". If using bool, there has to be True or False."
            ) from err

        if parsed_args[1]:
            raise RuntimeError(
                f"Config args parsing failed on unknown args: {parsed_args[1]}."
                "It may happen if variable not exists in config."
            )

        # Non empty command line args
        parser_args_dict = {}

        # All variables are parsed as strings
        # If it should not be string, infer type
        for i, j in parsed_args[0].__dict__.items():

            if j is None:
                continue

            if self.is_structured:
                used_type = type(self.base_config_map[i])[i].allowed_types
            else:
                used_type = type(self)[i].allowed_types

            if used_type is not str:
                try:
                    # May fail if for example Litera["string1", "string2"]
                    parser_args_dict[i] = types.str_to_infer_type(j)
                except ValueError:
                    parser_args_dict[i] = j
            else:
                parser_args_dict[i] = j

        self.update(parser_args_dict)


class ConfigStructured(ConfigBase):
    """Class for creating config. Read why and how in config module docstrings."""

    is_structured = True

    def get_dict(self) -> dict:
        """Flatten data and return to string."""
        # From main class
        dict_of_values = super().get_dict()
        # From sub configs
        for i in vars(self).values():
            if hasattr(i, "myproperties_list"):
                subconfig_dict = i.get_dict()
                dict_of_values.update(subconfig_dict)

        return dict_of_values

    def propagatebase_config_map(self) -> None:
        """Provide transfering arguments from base or from sub configs.

        Config class has subconfigs. It is possible to access subconfigs attributes from main config or from
        any other level because of this function.
        """
        for i in vars(self).values():
            if hasattr(i, "myproperties_list"):
                for j in i.myproperties_list:
                    self.base_config_map[j] = i

            elif hasattr(i, "properties_list"):
                for j in i.properties_list:
                    self.base_config_map[j] = i

            if isinstance(i, ConfigBase):
                i.base_config_map = self.base_config_map

            if ConfigStructured in type(i).mro():
                i.propagatebase_config_map()  #  type: ignore



"""Module with some helpers for type hints and annotations."""

from mypythontools.types.type_conversions_internal import json_to_py, str_to_bool, str_to_infer_type
from mypythontools.types.type_hints_internal import (
    get_return_type_hints,
    small_validate,
    typechecked_compatible,
)
from mypythontools.types.types_internal import validate_sequence

__all__ = [
    "get_return_type_hints",
    "validate_sequence",
    "json_to_py",
    "small_validate",
    "str_to_bool",
    "str_to_infer_type",
    "typechecked_compatible",
]



"""Module with functions for 'plots' subpackage."""

from __future__ import annotations
from typing import TYPE_CHECKING

from typing_extensions import Literal

import mylogging

from .. import misc
from ..paths import PathLike, get_desktop_path

# Lazy imports
# from pathlib import Path

# import plotly as pl
# import matplotlib.pyplot as plt
# from IPython import get_ipython

if TYPE_CHECKING:
    import pandas as pd


class GetPlotlyLayouts:
    """Plotly configs for particular plot types.

    Use `fig.layout.update(get_plotly_layout.categorical_scatter(title="My title"))'
    """

    def general(self, title):
        """General layout used in various plots."""
        return {
            "title": {
                "text": title,
                "x": 0.5,
                "xanchor": "center",
                "yanchor": "top",
                "y": 0.9 if misc.GLOBAL_VARS.jupyter else 0.95,
            },
            "titlefont": {"size": 28},
            "font": {"size": 17},
        }

    def categorical_scatter(self, title):
        """If there are categories sharing an axis."""
        return {
            **self.general(title),
            "margin": {"l": 320, "r": 150, "b": 180, "t": 130},
            "titlefont": {"size": 34},
            "font": {"size": 22},
            "yaxis": {"tickfont": {"size": 18}, "ticksuffix": " ", "title": {"standoff": 20}},
        }

    def time_series(self, title, showlegend, yaxis="Values"):
        """Good for time series prediction."""
        return {
            **self.general(title),
            "yaxis": {"title": yaxis},
            "showlegend": showlegend,
            "legend_orientation": "h",
            "hoverlabel": {"namelength": -1},
            "margin": {"l": 160, "r": 130, "b": 160, "t": 110},
        }


get_plotly_layout = GetPlotlyLayouts()


def check_if_column_in_df(df: pd.DataFrame, column, parameter):
    """Check if column exists in DataFrame and if not, raise an KeyError."""
    if not column in df.columns:
        raise KeyError(
            f"Column {column} from parameter {parameter} not found in DataFrame. "
            f"Possible columns are: {df.columns}"
        )


def plot(
    df: pd.DataFrame,
    plot_library: Literal["plotly", "matplotlib"] = "plotly",
    title: str = "Plot",
    legend: bool = True,
    y_axis_name="Values",
    blue_column="",
    black_column="",
    grey_area: None | list[str] = None,
    save_path: None | PathLike = None,
    return_div: bool = False,
    show: bool = True,
) -> None | str:
    """Plots the data.

    Plotly or matplotlib can be used. It is possible to highlite two columns with different
    formating. It is usually used for time series visualization, but it can be used for different use case of
    course.

    Args:
        df (pd.DataFrame): Data to be plotted.
        plot_library (Literal['plotly', 'matplotlib'], optional): 'plotly' or 'matplotlib'
            Defaults to "plotly".
        legend (bool, optional): Whether display legend or not. Defaults to True.
        blue_column (str, optional): Column name that will be formatted differently (blue). Defaults to "".
        black_column (str, optional): Column name that will be formatted differently (black, wider).
            And is surrounded by grey_area. Can be empty (if grey_area is None). Defaults to "".
        grey_area (None | list[str]), optional): Whether to show grey area surrounding the black_column.
            Can be None, or list of ['lower_bound_column', 'upper_bound_column']. Both columns has to be
            in df. Defaults to None.
        save_path (None | PathLike, optional): Whether save the plot.  If False or "", do not save,
            if Path as str path, save to defined path. If "DESKTOP" save to desktop. Defaults to None.
        return_div (bool, optional): If True, return html div with plot as string. If False, just plot and
            do not return. Defaults to False.
        show (bool, optional): Can be evaluated, but not shown (testing reasons). Defaults to True.

    Returns:
        None | str: Only if return_div is True, else None.

    Examples:
        Plot dataframe with

        >>> import pandas as pd
        >>> df = pd.DataFrame([[None, None, 1], [None, None, 2], [3, 3, 6], [3, 2.5, 4]])
        >>> plot(df, show=False)  # Show False just for testing reasons

        Example of filling grey area between columns

        >>> df = pd.DataFrame(
        ...     [[None, None, 0], [None, None, 2], [3, 3, 3], [2, 5, 4]], columns=["0", "1", "2"]
        ... )
        >>> plot(df, grey_area=["0", "1"], show=False)

        You can use matplotlib

        >>> plot(df, plot_library="matplotlib")

    Raises:
        KeyError: If defined column not found in DataFrame
    """
    if save_path == "DESKTOP":
        save_path = get_desktop_path() / "plot.html"

    if plot_library == "matplotlib":
        if misc.GLOBAL_VARS.jupyter:
            from IPython import get_ipython

            get_ipython().run_line_magic("matplotlib", "inline")  # type: ignore

        import matplotlib.pyplot as plt
        from matplotlib import rcParams

        rcParams["figure.figsize"] = (12, 8)

        df.plot()
        if legend:
            plt.legend(
                loc="upper center",
                bbox_to_anchor=(0.5, 1.05),
                ncol=3,
                fancybox=True,
                shadow=True,
            )

        if save_path:
            plt.savefig(save_path)

        if show:
            plt.show()

    elif plot_library == "plotly":

        import plotly as pl

        if misc.GLOBAL_VARS.jupyter:
            pl.io.renderers.default = "notebook_connected"

        used_columns = list(df.columns)

        graph_data = []

        if grey_area:
            check_if_column_in_df(df, grey_area[0], "grey_area[0]")
            check_if_column_in_df(df, grey_area[1], "grey_area[1]")

            upper_bound = pl.graph_objs.Scatter(
                name="Upper bound",
                x=df.index,
                y=df[grey_area[1]],
                line={"width": 0},
            )

            used_columns.remove(grey_area[1])
            graph_data.append(upper_bound)

            lower_bound = pl.graph_objs.Scatter(
                name="Lower bound",
                x=df.index,
                y=df[grey_area[0]],
                line={"width": 0},
                fillcolor="rgba(68, 68, 68, 0.3)",
                fill="tonexty",
            )

            used_columns.remove(grey_area[0])
            graph_data.append(lower_bound)

        if black_column:
            check_if_column_in_df(df, black_column, "black_column")

            surrounded = pl.graph_objs.Scatter(
                name=black_column,
                x=df.index,
                y=df[black_column],
                line={
                    "color": "rgb(51, 19, 10)",
                    "width": 5,
                },
                fillcolor="rgba(68, 68, 68, 0.3)",
                fill="tonexty" if grey_area else None,
            )

            used_columns.remove(black_column)
            graph_data.append(surrounded)

        if blue_column:

            check_if_column_in_df(df, blue_column, "blue_column")

            blue_column_ax = pl.graph_objs.Scatter(
                name=str(blue_column),
                x=df.index,
                y=df[blue_column],
                line={
                    "color": "rgb(31, 119, 180)",
                    "width": 2,
                },
            )

            used_columns.remove(blue_column)
            graph_data.append(blue_column_ax)

        fig = pl.graph_objs.Figure(data=graph_data)

        for i in df.columns:
            if i in used_columns:

                # Can be multiindex
                name = " - ".join([str(j) for j in i]) if isinstance(i, tuple) else i

                fig.add_trace(pl.graph_objs.Scatter(x=df.index, y=df[i], name=name))

        fig.layout.update(get_plotly_layout.time_series(title, legend, y_axis_name))

        if show:
            fig.show()

        if save_path:
            fig.write_html(save_path)

        if return_div:

            fig.layout.update(
                title=None,
                height=290,
                paper_bgcolor="#d9f0e8",
                margin={"b": 35, "t": 35, "pad": 4},
            )

            return pl.offline.plot(
                fig,
                include_plotlyjs=False,
                output_type="div",
            )



"""Module with functions for 'misc' subpackage."""

from __future__ import annotations
from typing import Callable, Any, Iterable
import builtins
import time
import sys
from pathlib import Path
import os
import importlib.util
import subprocess

import pandas as pd
from tabulate import tabulate

import mylogging

from ..paths import PathLike, validate_path


class Emojis:
    """Emojis that can be printed."""

    PARTY = "🎉"  # "\U0001F389"
    DISAPPOINTMENT = " ☹️ "  # "\U0001F641"


EMOJIS = Emojis()


class GlobalVars:
    """Global variables that can be useful."""

    @property
    def jupyter(self):
        """If runs in Jupyter, it returns True."""
        return True if hasattr(builtins, "__IPYTHON__") else False

    @property
    def is_tested(self):
        """If is tested with Pytest, it returns True."""
        return True if "PYTEST_CURRENT_TEST" in os.environ else False


GLOBAL_VARS = GlobalVars()

DEFAULT_TABLE_FORMAT = {
    "tablefmt": "grid",
    "floatfmt": ".3f",
    "numalign": "center",
    "stralign": "center",
}


class TimeTable:
    """Class that create printable table with spent time on various phases that runs sequentionally.

    Add entry when current phase end (not when it starts).

    Example:
        >>> import time
        ...
        >>> time_table = TimeTable()
        >>> time.sleep(0.01)
        >>> time_table.add_entry("First phase")
        >>> time.sleep(0.02)
        >>> time_table.add_entry("Second phase")
        >>> time_table.add_entry("Third phase")
        >>> time_table.finish_table()
        ...
        >>> print(time_table.time_table)
        +--------------+--------------+
        |     Time     |  Phase name  |
        +==============+==============+
        | First phase  |    0...
    """

    def __init__(self) -> None:
        """Init the table."""
        self.time_df: pd.DataFrame = pd.DataFrame()
        self.time_table: str = ""
        self.times: list[tuple[str, float]] = []
        self.last_time: float = time.time()

    def add_entry(self, phase_name: str) -> None:
        """Add new line to the Time table."""
        self.times.append((phase_name, round((time.time() - self.last_time), 3)))

    def finish_table(self, table_format: None | dict = None) -> None:
        """Create time table.

        Args:
            table_format (None | dict, optional): Dict of format settings used in tabulate. If None, default
                DEFAULT_TABLE_FORMAT is used. Defaults to None.
        """
        if not table_format:
            table_format = DEFAULT_TABLE_FORMAT

        self.add_entry("Completed")
        self.time_df = pd.DataFrame(self.times, columns=["Time", "Phase name"])
        self.time_table = tabulate(self.time_df.values, headers=list(self.time_df.columns), **table_format)


def check_library_is_available(name, message="default"):
    """Make one-liner for checking whether some library is installed.

    If running on venv, it checks only this venv, no global site packages.

    Args:
        name (str): Name of the library.
        message (str, optional): Message that will be printed when library not installed. Defaults to "default".

    Raises:
        ModuleNotFoundError: If module is installed, error is raised.

    Example:
        >>> check_library_is_available("typing_extensions")
        >>> check_library_is_available("not_installed_lib")
        Traceback (most recent call last):
        ModuleNotFoundError: ...
    """
    if message == "default":
        message = (
            f"Library {name} is necessary and not available. Some libraries are used in just for small"
            f"part of module, so not installed by default. Use \n\n\tpip install {name}\n\n"
        )

    if not importlib.util.find_spec(name):
        raise ModuleNotFoundError(message)


def check_script_is_available(name, install_library=None, message="default"):
    """Check if python script is available.

    This doesn't need to be installed in current venv, but anywhere on computer.

    Args:
        name (str): Name of the script. E.g "black'.
        install_library (str, optional): Install script with this library added to default message.
            Defaults to None.
        message (str, optional): Message that will be printed when library not installed.
            Defaults to "default".

    Raises:
        RuntimeError: If module is installed, error is raised.

    Example:
        >>> check_script_is_available("black")
        >>> check_script_is_available("not_existing_script")
        Traceback (most recent call last):
        RuntimeError: ...
    """
    if message == "default":
        message = f"Python script {name} is necessary and not available. "

    if install_library:
        message = message + f"To get this executable available, do \n\n\tpip install {name}\n\n"

    exists = False

    for i in ["--version", "--help", ""]:
        try:

            result = subprocess.run(
                f"{name} {i}", check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, shell=True
            )

            if result.returncode == 0:
                exists = True

            break

        except Exception:
            pass

    if not exists:
        raise RuntimeError(mylogging.format_str(message))


def watchdog(timeout: int | float, function: Callable, *args, **kwargs) -> Any:
    """Time-limited execution for python function. TimeoutError raised if not finished during defined time.

    Args:
        timeout (int | float): Max time execution in seconds.
        function (Callable): Function that will be evaluated.
        *args: Args for the function.
        *kwargs: Kwargs for the function.

    Raises:
        TimeoutError: If defined time runs out.
        RuntimeError: If function call with defined params fails.

    Returns:
        Any: Depends on used function.

    Examples:
        >>> import time
        >>> def sleep(sec):
        ...     for _ in range(sec):
        ...         time.sleep(1)
        >>> watchdog(1, sleep, 0)
        >>> watchdog(1, sleep, 10)
        Traceback (most recent call last):
        TimeoutError: ...
    """
    old_tracer = sys.gettrace()

    def tracer(frame, event, arg, start=time.time()):
        """Sys trace helpers that checks the time for watchdog."""
        now = time.time()
        if now > start + timeout:
            raise TimeoutError("Time exceeded")
        return tracer if event == "call" else None

    try:
        sys.settrace(tracer)
        result = function(*args, **kwargs)

    except TimeoutError:
        sys.settrace(old_tracer)
        raise TimeoutError(
            "Timeout defined in watchdog exceeded.",
        )

    except Exception:
        sys.settrace(old_tracer)
        raise RuntimeError(
            f"Watchdog with function {function.__name__}, args {args} and kwargs {kwargs} failed."
        )

    finally:
        sys.settrace(old_tracer)

    return result


def delete_files(paths: PathLike | Iterable[PathLike]):
    """Delete file or Sequence of files. If no permissions or file is open, it passes without error."""
    if isinstance(paths, (Path, str, os.PathLike)):
        paths = [paths]

    for i in paths:
        try:
            validate_path(i).unlink()
        except (FileNotFoundError, OSError):
            pass


def print_progress(name: str, verbose: bool = True):
    """Print current step of some process.

    Divide it with newlines so it's more readable.

    Args:
        name (str): Name current step.
        verbose (bool): It is possible to turn off logging of progress with one parameter config value.
            Defaults to True.
    """
    if verbose:
        print(f"\n{name}\n")



from pathlib import Path
import sys

sys.path.insert(0, Path(__file__).parents[2].as_posix())

from mypythontools.config import MyProperty, ConfigBase, ConfigStructured

if __name__ == "__main__":

    class SimpleConfig(ConfigStructured):
        def __init__(self) -> None:
            self.simple_sub_config = self.SimpleSubConfig()

        class SimpleSubConfig(ConfigBase):
            @MyProperty
            def bool_arg() -> bool:
                """This should be in CLI help."""
                return False

            @MyProperty
            def int_arg() -> int:
                """This should be in CLI help."""
                return 123

            @MyProperty
            def float_arg() -> float:
                return 123

            @MyProperty
            def str_arg() -> str:
                return "123"

            @MyProperty
            def list_arg() -> list:
                return []

            @MyProperty
            def dict_arg() -> dict:
                return {}

        @MyProperty
        def on_root() -> dict:
            """jes"""
            return {}

    config = SimpleConfig()

    config.with_argparse("How it works.")

    for i, j in config.get_dict().items():
        if "666" in str(j) or j is True:
            print(i, j, type(j))



"""
Module with miscellaneous functions that do not fit into other subpackage but are not big enough have it's own
subpackage.
"""

from mypythontools.misc.misc_internal import (
    check_library_is_available,
    check_script_is_available,
    DEFAULT_TABLE_FORMAT,
    delete_files,
    EMOJIS,
    GLOBAL_VARS,
    print_progress,
    TimeTable,
    watchdog,
)

__all__ = [
    "check_library_is_available",
    "check_script_is_available",
    "DEFAULT_TABLE_FORMAT",
    "delete_files",
    "EMOJIS",
    "GLOBAL_VARS",
    "print_progress",
    "TimeTable",
    "watchdog",
]



"""Runs before every pytest test. Used automatically (at least at VS Code)."""
from __future__ import annotations
import os
from pathlib import Path
import sys
import pytest


# Find paths and add to sys.path to be able to use local version and not installed mypythontools version
root = Path(__file__).parent

if root not in sys.path:
    sys.path.insert(0, root.as_posix())

import mypythontools_cicd as cicd
import mypythontools

cicd.tests.setup_tests(matplotlib_test_backend=True)

# Can be loaded from tests here or tests in test project
test_project_path = (
    Path("tests").resolve() / "tested project" if Path.cwd().name != "tested project" else Path.cwd()
)


@pytest.fixture(autouse=True)
def setup_tests():
    """Configure tests. Runs automatically from pytest and is called if running from file."""
    cwd_backup = Path.cwd()

    os.chdir(test_project_path.as_posix())
    mypythontools.paths.PROJECT_PATHS.reset_paths()

    yield

    os.chdir(cwd_backup.as_posix())



"""Module with functions for 'property' subpackage."""

from __future__ import annotations

from typing import Generic, TypeVar, Callable, Type, overload, Any

from typeguard import check_type

from .. import types


T = TypeVar("T")
U = TypeVar("U")


# Needs to inherit from property to be able to use help tooltip
class MyPropertyClass(property, Generic[T]):
    """Python property on steroids. Check module docstrings for more info."""

    # Property is inherited just for formatting help in IDE, so not called from init
    def __init__(self, fget: Callable[..., T] = None, doc=None):  # pylint: disable=super-init-not-called
        """Init property."""
        if fget:
            self.allowed_types = types.get_return_type_hints(fget)

            self.init_function = fget

            if doc:
                self.__doc__ = doc
            elif fget.__doc__:
                self.__doc__ = fget.__doc__
            else:
                self.__doc__ = None

        self.public_name = ""
        self.private_name = ""

    def default_fset(self, used_object, content) -> None:
        """Define how new values will be stored."""
        setattr(used_object, self.private_name, content)

    def __set_name__(self, _, name):
        """Define names. Private usually with underscore."""
        self.public_name = name
        self.private_name = "_" + name

    @overload
    def __get__(self, used_object: None, objtype: Any = None) -> MyPropertyClass[T]:
        ...

    @overload
    def __get__(self, used_object: U, objtype: Type[U] = None) -> T:
        ...

    def __get__(self, used_object, objtype=None):
        """Define what happens when you want access config attribute.

        If used on class, class itself is returned.
        """
        if not used_object:
            return self

        # Expected value can be nominal value or function, that return that value
        content = getattr(used_object, self.private_name)
        if isinstance(content, staticmethod):
            content = content.__func__

        if callable(content):
            # Depends whether it's staticmethod or not
            try:
                value = content(used_object)
            except TypeError:
                value = content()

        else:
            value = content

        return value

    def __set__(self, used_object, content: T | Callable[..., T]):
        """Define what happen if user set new config value."""
        # You can setup value or function, that return that value
        if callable(content):
            result = content(used_object)
        else:
            result = content

        if self.allowed_types:
            check_type(expected_type=self.allowed_types, value=result, argname=self.public_name)

        self.default_fset(used_object, result)


# Define as static method as it can be used directly
def init_my_properties(self):
    """Init private values of properties.

    Property usually get value of some private variable. This leads to many unnecessary code sometimes.
    """
    if not hasattr(self, "myproperties_list"):
        setattr(self, "myproperties_list", [])

    for i in vars(type(self)).values():
        if isinstance(i, MyPropertyClass):
            self.myproperties_list.append(i.public_name)
            setattr(
                self,
                i.private_name,
                i.init_function,
            )


def MyProperty(f: Callable[..., T]) -> MyPropertyClass[T]:  # pylint: disable=invalid-name
    """Wrap MyProperty to work as expected.

    If not using this workaround, but use class decorator, IDE complains that property has no defined
    setter. It's also prepended with staticmethod so IDE don't complain about On the other hand, it use
    correct type hint.

    Args:
        f(Callable[..., T]): Usually it's decorated method from class. It can be static as well.
    """
    return MyPropertyClass[T](f)


# TODO - Use PEP 614 and define type just i n class decorator
# Python 3.9 necessary



"""Module with functions for 'paths' subpackage."""

from __future__ import annotations
from typing import Sequence, Union
from pathlib import Path
import sys
import builtins

import mylogging

from ..types import validate_sequence

PathLike = Union[Path, str]  # Path is included in PathLike
"""Str pr pathlib Path. It can be also relative to current working directory."""


class ProjectPaths:
    """Define paths for usual python projects like root path, docs path, init path etc.

    You can find paths, that are lazy evaluated only after you ask for them. They are inferred
    automatically, but if you have alternative structure, you can also set it. Getters return path objects,
    so it's posix.

    Note:
        If you use paths in `sys.path.insert` or as subprocess main parameter, do not forget to convert it
        to string with `as_posix()`.
    """

    def __init__(self) -> None:
        """Init the paths."""
        self._root = None
        self._app = None
        self._init = None
        self._tests = None
        self._docs = None
        self._readme = None

    def add_root_to_sys_path(self) -> None:
        """As name suggest, add root to sys.paths on index 0."""
        if self.root.as_posix() not in sys.path:
            sys.path.insert(0, self.root.as_posix())

    @property
    def root(self) -> Path:
        """Path where all project is (docs, tests...). Root is usually current working directory.

        Type:
            Path

        Default:
            Path.cwd()
        """
        # If not value yet, set it first
        if not self._root:
            new_root_path = Path.cwd()

            # If using jupyter notebook from tests - very specific use case
            if new_root_path.name == "tests" and hasattr(builtins, "__IPYTHON__"):
                new_root_path = new_root_path.parent

            self._root = new_root_path

        return self._root

    @root.setter
    def root(self, new_path: PathLike) -> None:
        self._root = validate_path(new_path)

    @property
    def init(self) -> Path:
        """Path to __init__.py.

        Type:
            Path

        Default:
            **/__init__.py
        """
        if not self._init:
            exclude = []
            for i in ["docs", "tests"]:
                try:
                    exclude.append(getattr(self, i))
                except AttributeError:
                    pass

            self._init = find_path(
                "__init__.py",
                self.root,
                exclude_paths=exclude,
            )

        return self._init

    @init.setter
    def init(self, new_path: PathLike) -> None:
        self._init = validate_path(new_path)

    @property
    def app(self) -> Path:
        """Folder where python scripts are (and __init__.py).

        Type:
            Path

        Default:
            __App_path
        """
        if not self._app:
            self._app = self.init.parent

        return self._app

    @app.setter
    def app(self, new_path: PathLike) -> None:
        self._app = validate_path(new_path)

    @property
    def tests(self) -> Path:
        """Folder where tests are stored. Usually root / tests.

        'test', 'Test', 'Tests', 'TEST', 'TESTS' also inferred if on root.

        Type:
            Path

        Default:
            root_path/tests
        """
        if not self._tests:
            for i in ["tests", "test", "Test", "Tests", "TEST", "TESTS"]:
                if (self.root / i).exists():
                    self._tests = self.root / i
                    return self._tests

            raise RuntimeError("Test path not found.")

        return self._tests

    @tests.setter
    def tests(self, new_path: PathLike) -> None:
        self._tests = validate_path(new_path)

    @property
    def docs(self) -> Path:
        """Where documentation is stored. Usually root / docs.

        'doc', 'Doc', 'Docs', 'DOC', 'DOCS' also inferred if on root.

        Type:
            Path

        Default:
            root_path/docs
        """
        if not self._docs:
            for i in ["docs", "doc", "Doc", "Docs", "DOC", "DOCS"]:
                if (self.root / i).exists():
                    self._docs = self.root / i
                    return self._docs

            raise RuntimeError("Test path not found.")

        return self._docs

    @docs.setter
    def docs(self, new_path: PathLike) -> None:
        self._docs = validate_path(new_path)

    @property
    def readme(self) -> Path:
        """Return README path whether it's capitalized or not.

        'Readme.md', 'readme.md', and rst extension also inferred if on root.

        Type:
            Path

        Default:
            root_path/README.md
        """
        if not self._readme:
            for i in ["README.md", "Readme.md", "readme.md", "README.rst", "Readme.rst", "readme.rst"]:
                if (self.root / i).exists():
                    self._readme = self.root / i
                    return self._readme
            raise RuntimeError("Readme path not found.")

        return self._readme

    @readme.setter
    def readme(self, new_path: PathLike) -> None:
        self._readme = validate_path(new_path)

    def reset_paths(self):
        """Reset all the paths to default."""
        self._root = None
        self._app = None
        self._init = None
        self._tests = None
        self._docs = None
        self._readme = None


PROJECT_PATHS = ProjectPaths()


def find_path(
    name: str,
    folder: PathLike | None = None,
    exclude_names: Sequence[str] = ("node_modules", "build", "dist"),
    exclude_paths: Sequence[PathLike] = (),
    levels: int = 5,
) -> Path:
    """Search for file or folder in defined folder (cwd() by default) and return it's path.

    Args:
        name (str): Name of folder or file that should be found. If using file, use it with extension
            e.g. "app.py".
        folder (PathLike | None, optional): Where to search. If None, then root is used (cwd by default).
            Defaults to None.
        exclude_names (Sequence[str], optional): List or tuple of ignored names. If this name is whenever in
            path, it will be ignored. Defaults to ('node_modules', 'build', 'dist').
        exclude_paths (Sequence[PathLike], optional): List or tuple of ignored paths. If defined path is
            subpath of found file, it will be ignored. If relative, it has to be from cwd. Defaults to ().
        levels (str, optional): Recursive number of analyzed folders. Defaults to 5.

    Returns:
        Path: Found path.

    Raises:
        FileNotFoundError: If file is not found.
    """
    validate_sequence(exclude_names, "exclude_names")
    validate_sequence(exclude_paths, "exclude_paths")

    folder = PROJECT_PATHS.root if not folder else validate_path(folder)

    for lev in range(levels):
        glob_file_str = f"{'*/' * lev}{name}"

        for i in folder.glob(glob_file_str):
            is_wanted_file = True
            for j in exclude_names:
                if j in i.parts:
                    is_wanted_file = False
                    break

            if is_wanted_file:
                for j in exclude_paths:
                    excluded_name = Path(j).resolve()
                    if i.as_posix().startswith(excluded_name.as_posix()):
                        is_wanted_file = False
                        break

            if is_wanted_file:
                return i

    # If not returned - not found
    raise FileNotFoundError(f"File `{name}` not found")


def get_desktop_path() -> Path:
    """Get desktop path.

    Returns:
        Path: Return pathlib Path object. If you want string, use `.as_posix()`

    Example:
        >>> desktop_path = get_desktop_path()
        >>> desktop_path.exists()
        True
    """
    return Path.home() / "Desktop"


def validate_path(path: PathLike) -> Path:
    """Convert to pathlib path, resolve to full path and check if exists.

    Args:
        path (PathLike): Validated path.

    Raises:
        FileNotFoundError: If file do not exists.

    Returns:
        Path: Pathlib Path object.

    Example:
        >>> from pathlib import Path
        >>> existing_path = validate_path(Path.cwd())
        >>> non_existing_path = validate_path("not_existing")
        Traceback (most recent call last):
        FileNotFoundError: ...
    """
    path = Path(path).resolve()
    if not path.exists():
        raise FileNotFoundError(f"File nor folder found on defined path {path}")
    return path



"""Module contains MyProperty class that is alternative to normal python property. It's implemented via
descriptor and edited `__get__` and `__set__` magic methods. 

There is default setter, it's possible to auto init values on class init and values in setter can be
validated. This result in much less code written when using a lot of similar properties.

First call is lazy evaluated during first call.

Example of how can it be used is in module config.

Examples:
=========
    >>> from typing_extensions import Literal
    ...
    >>> class Example:
    ...     def __init__(self) -> None:
    ...         init_my_properties(self)
    ...
    ...     @MyProperty
    ...     @staticmethod  # Use staticmethod or add self or add pylint ignore message
    ...     def var() -> int:  # Type hints are validated.
    ...         '''
    ...         Type:
    ...             int
    ...
    ...         Default:
    ...             123
    ...
    ...         This is docstrings (also visible in IDE, because not defined dynamically).
    ...         Also visible in Sphinx documentation.'''
    ...
    ...         return 123  # This is initial value that can be edited.
    ...
    ...     @MyProperty
    ...     def var_literal(self) -> Literal[1, 2, 3]:  # Literal options are also validated
    ...         return 2
    ...
    ...     @MyProperty
    ...     def evaluated(self) -> int:  # If other defined value is change, computed property is also updated
    ...         return self.var + 1
    ...
    >>> config = Example()
    >>> config.var
    123
    >>> config.var = 665
    >>> config.var
    665
    >>> config.var = "String is problem"  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    TypeError: ...
    ...
    >>> config.var_literal = 4  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    TypeError: ...
    ...
    >>> config.evaluated
    666
    
    You can still setup a function (or lambda expression) as a new value
    and returned value still will be validated

    >>> config.var = lambda self: self.var_literal + 1
"""

from mypythontools.property.property_internal import MyProperty, init_my_properties

__all__ = ["MyProperty", "init_my_properties"]



"""Push the CI pipeline. Format, create commit from all the changes, push and deploy to PyPi."""



# TODO delete me
import sys
from pathlib import Path
sys.path.insert(0, (Path.cwd().parent / 'mypythontools_cicd').as_posix())
import mypythontools_cicd

from mypythontools_cicd.project_utils import project_utils_pipeline

if __name__ == "__main__":
    # All the parameters can be overwritten via CLI args
    project_utils_pipeline(
        reformat=True,
        test=True,
        test_options={"virtualenvs": ["venv/37", "venv/310"]},
        version="increment",
        docs=True,
        sync_requirements=False,
        commit_and_push_git=True,
        commit_message="New commit",
        tag="__version__",
        tag_message="New version",
        deploy=True,
        allowed_branches=("master", "main"),
    )

    # project_utils_pipeline(
    #     do_only="deploy",
    # )



"""Module with functions for 'terminal' subpackage."""

from __future__ import annotations
from pathlib import Path
import subprocess
import platform

from ..paths import PathLike


class TerminalCommandError(Exception):
    pass


SHELL_AND = " && " if platform.system() == "Windows" else " && "
"""If using some nonstandard shell can be edited here globally ';' can be used if && not working."""
EXECUTABLE = None if platform.system() == "Windows" else "/bin/bash"
"""To be able to use 'source' it points to '/bin/bash' on linux."""
PYTHON = "python" if platform.system() == "Windows" else "python3"


def terminal_do_command(
    command: str,
    shell: bool = True,
    cwd: None | PathLike = None,
    verbose: bool = True,
    error_header: str = "",
):
    """Run command in terminall and process output.

    Args:
        command (str): Command to run.
        shell (bool, optional): Same meaning as in ``subprocess.run()``. Defaults to False.
        cwd (None | PathLike, optional): Same meaning as in ``subprocess.run()``. Defaults to None.
        verbose (bool, optional): Whether print output to console. Defaults to True.
        error_header (str, optional): If meet error, message at the beginning of message. Default to "".

    Raises:
        RuntimeError: When process fails to finish or return non zero return code.
    """
    error = None

    try:
        result = subprocess.run(command, shell=shell, cwd=cwd, capture_output=True)
        if result.returncode == 0:
            if verbose:
                print(result.stdout.decode().strip("\r\n"))
        else:
            stderr = result.stderr.decode().strip("\r\n")
            stdout = result.stdout.decode().strip("\r\n")
            error = f"\n\nstderr:\n\n{stderr}\n\nstdout:\n\n{stdout}\n\n"

    except Exception:  # pylint: disable=broad-except
        error = "Suprocess command crashed internally in subprocess and did not finished."

    if error:
        header = f"{error_header}\n\n" if error_header else ""
        cwd_str = "on your project root" if cwd is None else f"in '{cwd}' folder"

        raise TerminalCommandError(
            f"{header}"
            f"Running command in terminal failed. Try command below in the terminal {cwd_str} "
            f"\n\n{command}\n\n"
            "On windows use cmd so script paths resolved correctly. Try it with administrator rights in\n"
            "your project root folder. Permission error may be one of the typical issue or some\n"
            "necessary library missing or installed elsewhere than in used venv.\n\n"
            f"Captured error: {error}",
        )


def get_console_str_with_quotes(string: PathLike):
    """In terminal if value or contain spaces, it's not taken as one param.

    This wraps it with quotes to be able to use paths and values as needed. Alternative to this function is to
    use python shlex library, list of commands and 'shlex.join' to get the command string.

    Args:
        string (str, Path): String  to be edited.

    Returns:
        str: Wrapped string that can be used in terminal.

    Example:
        >>> get_console_str_with_quotes("/path to file/file")
        '"/path to file/file"'
    """
    if isinstance(string, (Path)):
        string = string.as_posix()
    if not isinstance(string, str):
        string = str(string)
    string = string.strip("'")
    string = string.strip('"')
    return f'"{string}"'



"""Module with functions for 'type_hints' subpackage."""

from __future__ import annotations
import sys

# Import can be used in eval
from typing import (
    Any,
    Callable,
    Union,
    List,
    Dict,
    Tuple,
    Sequence,
    Iterable,
)  # pylint: disable=unused-import

from typing_extensions import Literal, get_origin, get_args, get_type_hints

from typeguard import typechecked  # , check_type

import mylogging


def typechecked_compatible(function):
    """Turns off type checking for old incompatible python versions.

    Mainly for new syntax like list[str] which raise TypeError.
    """

    # def decorator(func):
    #     if sys.version_info.minor < 9:
    #         return func
    #     return typechecked(func)

    if sys.version_info.minor < 9:
        return function
    return typechecked(function)


def get_return_type_hints(func: Callable) -> Any:
    """Return function return types.

    This is because `get_type_hints` result in error for some types in older
    versions of python and also that `__annotations__` contains only string, not types.

    Note:
        Sometimes it may use eval as literal_eval cannot use users globals so types like pd.DataFrame would
        fail. Therefore do not use it for evaluating types of users input for sake of security.

    Args:
        func (Callable): Function with type hints.

    Returns:
        Any: Type of return.

    Example:
        >>> # You can use Union as well as Literal
        >>> def union_return() -> int | float:
        ...     return 1
        >>> inferred_type = get_return_type_hints(union_return)
        >>> 'int' in str(inferred_type) and 'float' in str(inferred_type)
        True
        >>> def literal_return() -> Literal[1, 2, 3]:
        ...     return 1
        >>> inferred_type = get_return_type_hints(literal_return)
        >>> 'Literal' in str(inferred_type)
        True
    """
    if isinstance(func, staticmethod):
        func = func.__func__

    try:
        types = get_type_hints(func).get("return")
    except Exception:
        types = func.__annotations__.get("return")

    if isinstance(types, str) and "Union" in types:
        types = eval(types, func.__globals__)

    # If Union operator |, e.g. int | str - get_type_hints() result in TypeError
    # Convert it to Union
    elif isinstance(types, str) and "|" in types:
        str_types = [i.strip() for i in types.split("|")]
        for i, j in enumerate(str_types):
            for k in ["list", "dict", "tuple"]:
                if k in j:
                    str_types[i] = j.replace(k, k.capitalize())
        try:
            evaluated_types = [eval(i, {**globals(), **func.__globals__}) for i in str_types]
        except Exception:
            raise RuntimeError("Evaluating of function return type failed. Try it on python 3.9+.")

        types = Union[evaluated_types[0], evaluated_types[1]]  # type: ignore

        if len(evaluated_types) > 2:
            for i in evaluated_types[2:]:
                types = Union[types, i]

    return types


class ValidationError(TypeError):
    """To know that error is because of bad config type and not some TypeError from the inside."""

    pass


# def validate(value, allowed_type: Any, name: str) -> None:
#     """Type validation. It also works for Union and validate Literal values.

#     Instead of typeguard validation, it define just subset of types, but is simplier
#     and needs no extra import, therefore can be faster.

#     Args:
#         value (Any): Value that will be validated.
#         allowed_type (Any, optional): For example int, str or list. It can be also Union
#             or Literal. If Literal, validated value has to be one of Literal values.
#             Defaults to None.
#         name (str | None, optional): If error raised, name will be printed. Defaults to None.

#     Raises:
#         ValidationError: Type does not fit.

#     Examples:
#         >>> from typing_extensions import Literal
#         ...
#         >>> validate(1, int)
#         >>> validate(None, list | None)
#         >>> validate("two", Literal["one", "two"])
#         >>> validate("three", Literal["one", "two"])
#         Traceback (most recent call last):
#         ValidationError: ...
#     """
#     check_type(value=value, expected_type=allowed_type, argname=name)

# TODO Wrap error with colors and remove stack only to configuration line...
# try:
#     check_type(value=value, expected_type=allowed_type, argname=name)
# except TypeError:

#     # ValidationError(mylogging.format_str("validate"))

#     raise


def small_validate(value, allowed_type: None | Any = None, name: str | None = None) -> None:
    """Type validation. It also works for Union and validate Literal values.

    Instead of typeguard validation, it define just subset of types, but is simplier
    and needs no extra import, therefore can be faster.

    Args:
        value (Any): Value that will be validated.
        allowed_type (Any, optional): For example int, str or list. It can be also Union or Literal.
            If Literal, validated value has to be one of Literal values. If None, it's skipped.
            Defaults to None.
        name (str | None, optional): If error raised, name will be printed. Defaults to None.

    Raises:
        TypeError: Type does not fit.

    Examples:
        >>> from typing_extensions import Literal
        ...
        >>> small_validate(1, int)
        >>> small_validate(None, Union[list, None])
        >>> small_validate("two", Literal["one", "two"])
        >>> small_validate("three", Literal["one", "two"])  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ValidationError: ...
    """
    if allowed_type:
        # If Union
        if get_origin(allowed_type) == Union:

            if type(value) in get_args(allowed_type):
                return
            else:
                raise ValidationError(
                    mylogging.format_str(
                        f"Allowed type for variable '{name}' are {allowed_type}, but you try to set an {type(value)}"
                    )
                )

        # If Literal - parse options
        elif get_origin(allowed_type) == Literal:
            options = getattr(allowed_type, "__args__")
            if value in options:
                return
            else:
                raise ValidationError(
                    f"New value < {value} > for variable < {name} > is not in allowed options {options}."
                )

        else:
            if isinstance(value, allowed_type):  # type: ignore
                return
            else:
                raise ValidationError(
                    f"Allowed allowed_type for variable < {name} > is {allowed_type}, but you try to set an {type(value)}"
                )



"""Wrap python subprocess with some extra stuff

You can find here formatting errors, resolving venv script paths, unifying syntax for linux and windows etc.
Main function is 'terminal_do_command'. Use strings as input and if there can be space in some command, use
'get_console_str_with_quotes'.
"""
from mypythontools.terminal.terminal_internal import (
    EXECUTABLE,
    get_console_str_with_quotes,
    PYTHON,
    SHELL_AND,
    terminal_do_command,
)

__all__ = [
    "EXECUTABLE",
    "get_console_str_with_quotes",
    "PYTHON",
    "SHELL_AND",
    "terminal_do_command",
]



"""Module with functions for 'types' subpackage."""


def validate_sequence(value, variable):
    if isinstance(value, str):
        raise TypeError(
            f"Variable '{variable}' must not be string, but Sequence. It can be tuple or list for example. "
            "Beware that if you use tuple with just one member like this ('string'), it's finally parsed as "
            "string. If this is the case, add coma like this ('string',)."
        )



"""Module with functions for 'types' subpackage."""

from __future__ import annotations
from typing import Callable, Any, Union, Iterable
import ast


def str_to_infer_type(string_var: str) -> Any:
    """Convert string to another type (for example to int, float, list or dict).

    Args:
        string_var (str): String that should be converted.

    Returns:
        Any: New inferred type.

    Examples:
        >>> type(str_to_infer_type("1"))
        <class 'int'>
        >>> type(str_to_infer_type("1.2"))
        <class 'float'>
        >>> type(str_to_infer_type("['one']"))
        <class 'list'>
        >>> type(str_to_infer_type("{'one': 1}"))
        <class 'dict'>
    """
    return ast.literal_eval(string_var)


def json_to_py(json: dict, replace_comma_decimal: bool = True, replace_true_false: bool = True) -> Any:
    """Take json and eval it from strings. If string to string, if float to float, if object then to dict.

    When to use? - If sending object as parameter in function.

    Args:
        json (dict): JSON with various formats as string.
        replace_comma_decimal (bool, optional): Some countries use comma as decimal separator (e.g. 12,3).
            If True, comma replaced with dot (Only if there are no brackets (list, dict...)
            and if not converted to number string remain untouched) . For example '2,6' convert to 2.6.
            Defaults to True
        replace_true_false (bool, optional): If string is 'false' or 'true' (for example from javascript),
            it will be capitalized first for correct type conversion. Defaults to True

    Returns:
        dict: Python dictionary with correct types.

    Example:
        >>> # Can be beneficial for example when communicating with JavaScript
        >>> json_to_py({'one_two': '1,2'})
        {'one_two': 1.2}
    """
    import ast

    evaluated = json.copy()

    for i, j in json.items():

        replace_condition = isinstance(j, str) and "(" not in j and "[" not in j and "{" not in j

        if replace_comma_decimal and replace_condition:
            j = j.replace(",", ".")

        if replace_true_false and replace_condition:
            if j == "true":
                evaluated[i] = True
            if j == "false":
                evaluated[i] = False
            if j == "true" or j == "false":
                continue

        try:
            evaluated[i] = ast.literal_eval(j)
        except Exception:
            pass

    return evaluated


def str_to_bool(bool_str):
    """Convert string to bool. Usually used from argparse. Raise error if don't know what value.

    Possible values for True: 'yes', 'true', 't', 'y', '1'
    Possible values for False: 'no', 'false', 'f', 'n', '0'

    Args:
        bool_str (str):

    Raises:
        TypeError: If not one of bool values inferred, error is raised.

    Returns:
        bool: True or False

    Example:
        >>> str_to_bool("y")
        True

    Argparse example::

        parser = argparse.ArgumentParser()

        parser.add_argument(
            "--test",
            choices=(True, False),
            type=str_to_bool,
            nargs="?",
        )
    """
    if isinstance(bool_str, bool):
        return bool_str
    if bool_str.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif bool_str.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise TypeError("Boolean value expected.")



from setuptools import setup, find_packages
import pkg_resources
import mypythontools

version = mypythontools.__version__

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("requirements.txt") as f:
    my_requirements = [str(requirement) for requirement in pkg_resources.parse_requirements(f)]

if __name__ == "__main__":

    setup(
        author_email="malachovd@seznam.cz",
        author="Daniel Malachov",
        description="Some tools/functions/snippets used across projects.",
        include_package_data=True,
        install_requires=my_requirements,
        # extras_require={},
        license="mit",
        long_description_content_type="text/markdown",
        long_description=readme,
        name="mypythontools",
        packages=find_packages(exclude=("tests**",)),
        platforms="any",
        project_urls={
            "Documentation": "https://mypythontools.readthedocs.io/",
            "Homepage": "https://github.com/Malachov/mypythontools",
        },
        python_requires=">=3.7",
        url="https://github.com/Malachov/mypythontools",
        version=version,
        classifiers=[
            "Development Status :: 3 - Alpha",
            # "Development Status :: 4 - Beta
            # "Development Status :: 5 - Production/Stable",
            "Environment :: Other Environment",
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "License :: OSI Approved :: MIT License",
            "Natural Language :: English",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python",
            "Topic :: Software Development :: Libraries :: Application Frameworks",
            "Topic :: Software Development :: Libraries :: Python Modules",
        ],
    )
